class MissingContextVariableError(Exception):
    """
    Raised when template field cannot be rendered because variable used in it is not present in the context.
    """
    def __init__(self, field_name, variable):
        super().__init__(
            'Cannot render field "{}", because variable "{}" is not defined in the context.'.format(
                field_name, variable
            )
        )



class BaseDispatcher:
    """
    Base class for sending notification over a communication channel (e.g. e-mail, sms, push).
    """
    def dispatch(self, notification):
        """
        This method should implement actual sending of ``notification``.
        """
        raise NotImplementedError()  # pragma: no cover



# Generated by Django 2.2rc1 on 2019-03-20 11:24

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='Notification',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('is_read', models.BooleanField(default=False, verbose_name='is read')),
                ('is_triggered', models.BooleanField(default=False, verbose_name='is triggered')),
                ('recipient', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE,
                                                related_name='notifications', to=settings.AUTH_USER_MODEL,
                                                verbose_name='recipient')),
            ],
            options={
                'verbose_name': 'notification',
                'verbose_name_plural': 'notifications',
                'ordering': ('-created_at',),
            },
        ),
        migrations.CreateModel(
            name='NotificationTemplate',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('title', models.CharField(max_length=200, verbose_name='title')),
                ('text', models.TextField(blank=True, null=True, verbose_name='text')),
                ('trigger_action', models.CharField(blank=True, max_length=2500, null=True,
                                                    verbose_name='trigger action')),
                ('slug', models.SlugField(blank=True, max_length=200, null=True, unique=True, verbose_name='slug')),
            ],
            options={
                'verbose_name': 'notification template',
                'verbose_name_plural': 'notification templates',
            },
        ),
        migrations.CreateModel(
            name='NotificationRelatedObject',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('name', models.CharField(max_length=200, verbose_name='name')),
                ('object_id', models.TextField(db_index=True, verbose_name='object ID')),
                ('content_type', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL,
                                                   to='contenttypes.ContentType', verbose_name='content type')),
                ('notification', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE,
                                                   related_name='related_objects', to='pynotify.Notification',
                                                   verbose_name='notification')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='notification',
            name='template',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='notifications',
                                    to='pynotify.NotificationTemplate', verbose_name='template'),
        ),
    ]



from celery import shared_task

from .helpers import process_task


@shared_task
def notification_task(*args, **kwargs):
    """
    Celery task used in asynchronous mode. Just passes any arguments to ``process_task`` function.
    """
    process_task(*args, **kwargs)



# Generated by Django 2.1.5 on 2019-04-10 07:12

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pynotify', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='notification',
            name='extra_data',
            field=models.TextField(blank=True, null=True, verbose_name='extra data'),
        ),
    ]



# Generated by Django 3.0.2 on 2020-01-17 18:40

from django.db import migrations, models
import django.db.models.deletion


def convert_notification_templates_with_slug_to_admin_templates(apps, schema_editor):
    NotificationTemplate = apps.get_model("pynotify", "NotificationTemplate")
    AdminNotificationTemplate = apps.get_model("pynotify", "AdminNotificationTemplate")

    for notification_template in NotificationTemplate.objects.filter(slug__isnull=False):
        admin_template = AdminNotificationTemplate.objects.create(
            title=notification_template.title,
            text=notification_template.text,
            trigger_action=notification_template.trigger_action,
            slug=notification_template.slug,
        )
        notification_template.admin_template = admin_template
        notification_template.save()


class Migration(migrations.Migration):

    dependencies = [
        ('pynotify', '0003_related_object_name_not_required'),
    ]

    operations = [
        migrations.CreateModel(
            name='AdminNotificationTemplate',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('title', models.CharField(max_length=200, verbose_name='title')),
                ('text', models.TextField(blank=True, null=True, verbose_name='text')),
                ('trigger_action', models.CharField(blank=True, max_length=2500, null=True,
                                                    verbose_name='trigger action')),
                ('slug', models.SlugField(blank=False, max_length=200, null=False, unique=True, verbose_name='slug')),
            ],
            options={
                'verbose_name': 'admin notification template',
                'verbose_name_plural': 'admin notification templates',
            },
        ),
        migrations.AddField(
            model_name='notificationtemplate',
            name='admin_template',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL,
                                    related_name='notification_templates', to='pynotify.AdminNotificationTemplate',
                                    verbose_name='admin notification template'),
        ),
        migrations.RunPython(convert_notification_templates_with_slug_to_admin_templates),
        migrations.RemoveField(
            model_name='notificationtemplate',
            name='slug',
        ),
    ]



from django.dispatch import Signal

from .handlers import BaseHandler


notify_signal = Signal(providing_args=['recipients', 'title', 'text', 'trigger_action', 'related_objects',
                                       'extra_data', 'template_slug', 'dispatcher_classes'])


def notify(recipients, title=None, text=None, trigger_action=None, related_objects=None, extra_data=None,
           template_slug=None, dispatcher_classes=None):
    """
    Helper method to create a notification. Simply sends the ``notify_signal``.
    """
    if bool(template_slug) == (bool(title) | bool(text) | bool(trigger_action)):
        raise ValueError('Either provide template slug or template data, not both.')

    notify_signal.send(
        sender=None,
        recipients=recipients,
        title=title,
        text=text,
        trigger_action=trigger_action,
        related_objects=related_objects,
        extra_data=extra_data,
        template_slug=template_slug,
        dispatcher_classes=dispatcher_classes,
    )


class NotifyHandler(BaseHandler):
    """
    Notification handler for ``notify_signal``.
    """
    def get_recipients(self):
        return self.signal_kwargs['recipients']

    def get_template_data(self):
        return {x: self.signal_kwargs[x] for x in {'title', 'text', 'trigger_action'}}

    def get_related_objects(self):
        return self.signal_kwargs['related_objects']

    def get_extra_data(self):
        return self.signal_kwargs['extra_data']

    def get_template_slug(self):
        return self.signal_kwargs['template_slug']

    def get_dispatcher_classes(self):
        return self.signal_kwargs['dispatcher_classes'] or ()

    class Meta:
        signal = notify_signal



from django.conf import settings as django_settings


class Settings:
    """
    Holds default configuration values, the values can be overriden in settings with ``PYNOTIFY_`` prefix.
    """
    PREFIX = 'PYNOTIFY'
    DEFAULTS = {
        'AUTOLOAD_MODULES': None,
        'CELERY_TASK': 'pynotify.tasks.notification_task',
        'ENABLED': True,
        'RECEIVER': 'pynotify.receivers.SynchronousReceiver',
        'RELATED_OBJECTS_ALLOWED_ATTRIBUTES': {'get_absolute_url', },
        'STRIP_HTML': False,
        'TEMPLATE_CHECK': False,
        'TEMPLATE_PREFIX': '',
        'TEMPLATE_TRANSLATE': False,
    }

    def __getattr__(self, attr):
        if attr not in self.DEFAULTS:
            raise AttributeError('Invalid setting: "{}"').format(attr)

        return getattr(django_settings, '{}_{}'.format(self.PREFIX, attr), self.DEFAULTS[attr])


settings = Settings()



from collections import Iterable

from django.contrib.contenttypes.models import ContentType
from django.db.models import Model


class BaseSerializer:
    """
    Base class for serializing/deserializing signal kwargs. Its puprose is to transform signal kwargs to be directly
    JSON serializable (for compatible types, see https://docs.python.org/3/library/json.html#py-to-json-table).
    """
    def serialize(self, signal_kwargs):
        """
        This method should return serialized ``signal_kwargs``.
        """
        raise NotImplementedError()  # pragma: no cover

    def deserialize(self, signal_kwargs):
        """
        This method should return deserialized ``signal_kwargs``.
        """
        raise NotImplementedError()  # pragma: no cover


class ModelSerializer:
    """
    Serializes any model instance into its PK and ContentType PK and deserializes by fetching the model instance from
    database. Works recursively on nested dicts and iterables. Values that are not model instances are left intact.
    """
    CONTENT_TYPE_PK_KEY = 'ct_pk'
    INSTANCE_PK_KEY = 'obj_pk'

    def _is_serialized(self, value):
        return isinstance(value, dict) and list(value.keys()) == [self.CONTENT_TYPE_PK_KEY, self.INSTANCE_PK_KEY]

    def _is_deserialized(self, value):
        return isinstance(value, Model) and hasattr(value, 'pk')

    def _serialize(self, value):
        return {
            self.CONTENT_TYPE_PK_KEY: ContentType.objects.get_for_model(value).pk,
            self.INSTANCE_PK_KEY: value.pk,
        }

    def _deserialize(self, value):
        model_class = ContentType.objects.get_for_id(value[self.CONTENT_TYPE_PK_KEY]).model_class()
        return model_class.objects.get(pk=value[self.INSTANCE_PK_KEY])

    def _process(self, check_method, process_method, input):
        output = {}
        for key, value in input.items():
            if check_method(value):
                output[key] = process_method(value)
            elif isinstance(value, dict):
                output[key] = self._process(check_method, process_method, value)
            elif isinstance(value, Iterable) and not isinstance(value, str):
                output[key] = [process_method(i) if check_method(i) else i for i in value]
            else:
                output[key] = value

        return output

    def serialize(self, signal_kwargs):
        return self._process(self._is_deserialized, self._serialize, signal_kwargs)

    def deserialize(self, signal_kwargs):
        return self._process(self._is_serialized, self._deserialize, signal_kwargs)



from collections import Iterable

from django.core.exceptions import ImproperlyConfigured
from django.utils.functional import cached_property

from .helpers import register
from .models import AdminNotificationTemplate, Notification, NotificationTemplate


class HandlerMeta(type):
    """
    Registers handler for handling of signal defined in handler's ``Meta``.
    """
    def __new__(cls, name, bases, attrs):
        new_class = super().__new__(cls, name, bases, attrs)

        # register only descendants of BaseHandler
        if bases:
            if not getattr(new_class, 'Meta', None):
                raise ImproperlyConfigured('{} must have Meta class defined.'.format(name))

            if getattr(new_class.Meta, 'abstract', False):
                # delete the abstract attribute so it is not inherited by child classes
                delattr(new_class.Meta, 'abstract')
            else:
                if not getattr(new_class, 'Meta', None) or not getattr(new_class.Meta, 'signal', None):
                    raise ImproperlyConfigured('No signal defined in {}\'s Meta.'.format(name))

                if (
                    getattr(new_class.Meta, 'allowed_senders', None)
                    and not isinstance(new_class.Meta.allowed_senders, Iterable)
                ):
                    raise ImproperlyConfigured('{}\'s attribute "allowed_senders" must be iterable.'.format(name))

                register(
                    signal=new_class.Meta.signal,
                    handler_class=new_class,
                    allowed_senders=getattr(new_class.Meta, 'allowed_senders', None),
                )

        return new_class


class BaseHandler(metaclass=HandlerMeta):
    """
    Base class for handling creation of notification(s). Its purpose is to process signal kwargs sent over a defined
    signal. There should be typically one handler (inherited from this class) for each signal. The handler must define
    inner class ``Meta`` with following supported attributes:

        * ``signal``: Signal to which handler will be registered
        * ``allowed_senders``: Handler will be called only if signal was sent by allowed sender
        * ``abstract``: If set to True, the handler will not be registered

    Attributes:
        dispatcher_classes: An iterable of dispatcher classes that will be used to dispatch each notification.
        template_slug: Slug of an existing admin template to be used. If not defined, you must define
            ``get_template_data()`` method.
    """
    dispatcher_classes = ()
    template_slug = None

    @cached_property
    def _admin_template(self):
        template_slug = self.get_template_slug()
        if template_slug:
            return AdminNotificationTemplate.objects.get(slug=template_slug)
        else:
            return None

    @cached_property
    def _template(self):
        if self._admin_template:
            template, _ = NotificationTemplate.objects.get_or_create(
                title=self._admin_template.title,
                text=self._admin_template.text,
                trigger_action=self._admin_template.trigger_action,
                admin_template=self._admin_template,
            )
        else:
            template, _ = NotificationTemplate.objects.get_or_create(**self.get_template_data())

        return template

    def _init_dispatchers(self):
        self.dispatchers = []
        for dispatcher_class in set(self.get_dispatcher_classes()):
            self.dispatchers.append(self._init_dispatcher(dispatcher_class))

    def _init_dispatcher(self, dispatcher_class):
        return dispatcher_class()

    def _can_create_notification(self, recipient):
        """
        Returns ``True`` if notification can be created for ``recipient``.
        """
        return True

    def _create_notification(self, recipient):
        """
        Creates notification for ``recipient``.
        """
        if self._can_create_notification(recipient):
            return Notification.objects.create(
                recipient=recipient,
                template=self._template,
                related_objects=self.get_related_objects(),
                extra_data=self.get_extra_data(),
            )

    def _can_dispatch_notification(self, notification, dispatcher):
        """
        Returns ``True`` if ``notification`` can be dispatched using ``dispatcher``.
        """
        return True

    def _dispatch_notification(self, notification, dispatcher):
        """
        Dispatches ``notification`` using ``dispatcher``.
        """
        if self._can_dispatch_notification(notification, dispatcher):
            dispatcher.dispatch(notification)

    def _can_handle(self):
        """
        Returns ``True`` if handler should handle creating of notification(s).
        """
        return not self._admin_template or self._admin_template.is_active

    def handle(self, signal_kwargs):
        """
        Handles creation of notifications from ``signal_kwargs``.
        """
        self.signal_kwargs = signal_kwargs
        if self._can_handle():
            self._init_dispatchers()
            for recipient in self.get_recipients():
                notification = self._create_notification(recipient)

                if notification:
                    for dispatcher in self.dispatchers:
                        self._dispatch_notification(notification, dispatcher)

    def get_recipients(self):
        """
        Returns an iterable of recipients for which notification will be created.
        """
        raise NotImplementedError()  # pragma: no cover

    def get_template_data(self):
        """
        Returns kwargs used to create a template. Not called if template slug is used.
        """
        raise NotImplementedError()  # pragma: no cover

    def get_related_objects(self):
        """
        Returns a list or dictionary of related objects in format {"name": object}. Named related objects (i.e. those
        passed using a dictionary) can be referred in notification template.
        """
        return None

    def get_extra_data(self):
        """
        Returns a dictionary with extra data, the values must be JSON serializable.
        """
        return None

    def get_template_slug(self):
        """
        Returns slug of an admin template to be used.
        """
        return self.template_slug

    def get_dispatcher_classes(self):
        """
        Returns iterable of dispatcher classes used to dispatch notification(s).
        """
        return self.dispatcher_classes



from pydoc import locate

from django.core.exceptions import ImproperlyConfigured
from django.db.transaction import on_commit

from .config import settings
from .helpers import get_import_path
from .serializers import ModelSerializer


class BaseReceiver:
    """
    Base class for receiving signals. Its purpose is to pass signal kwargs to the notification handler.
    """
    def __init__(self, handler_class):
        self.handler_class = handler_class

    def receive(self, signal_kwargs):
        """
        This method should implement passing ``signal_kwargs`` to the handler.
        """
        raise NotImplementedError()  # pragma: no cover


class SynchronousReceiver(BaseReceiver):
    """
    Signal receiver that calls notification handler synchronously.
    """
    def __init__(self, handler_class):
        super().__init__(handler_class)
        self.handler = self.handler_class()

    def receive(self, signal_kwargs):
        self.handler.handle(signal_kwargs)


class AsynchronousReceiver(BaseReceiver):
    """
    Signal receiver that calls notification handler asynchronously via Celery.
    """
    def __init__(self, handler_class):
        super().__init__(handler_class)
        self.serializer_class = ModelSerializer
        self.celery_task = self._get_celery_task()

    def _get_celery_task(self):
        celery_task = settings.CELERY_TASK
        if not celery_task:
            raise ImproperlyConfigured(
                'CELERY_TASK setting must be set when using {}.'.format(self.__class__.__name__)
             )
        return locate(celery_task)

    def receive(self, signal_kwargs):
        # Call of the Celery task should be performed after current DB transaction is commited to avoid race condition,
        # e.g. accessing referenced object in the task before it has finished saving into DB.
        on_commit(lambda: self.celery_task.delay(
            get_import_path(self.handler_class),
            get_import_path(self.serializer_class),
            self.serializer_class().serialize(signal_kwargs),
        ))



from django.apps import AppConfig

from .helpers import autoload


class PyNotifyConfig(AppConfig):
    name = 'pynotify'
    verbose_name = 'PyNotify'

    def ready(self):
        from .notify import NotifyHandler  # noqa
        autoload()



# Generated by Django 3.0.2 on 2020-07-23 13:54

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pynotify', '0004_add_admin_template'),
    ]

    operations = [
        migrations.AddField(
            model_name='adminnotificationtemplate',
            name='is_active',
            field=models.BooleanField(default=True, verbose_name='is active'),
        ),
    ]



# Generated by Django 3.0.2 on 2020-01-17 17:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pynotify', '0002_notification_extra_data'),
    ]

    operations = [
        migrations.AlterField(
            model_name='notificationrelatedobject',
            name='name',
            field=models.CharField(blank=True, max_length=200, null=True, verbose_name='name'),
        ),
    ]



from importlib import import_module
import logging
from pydoc import locate

from bs4 import BeautifulSoup
from django.core.exceptions import ImproperlyConfigured
from django.utils.translation import ugettext as _

from .config import settings


logger = logging.getLogger(__name__)


class SecureRelatedObject:
    """
    Security proxy class allowing to access only string representation of the related object and a set of attributes
    defined in `RELATED_OBJECTS_ALLOWED_ATTRS` settings.
    """
    def __init__(self, related_object):
        self._object = related_object

    def __getattr__(self, name):
        if name in settings.RELATED_OBJECTS_ALLOWED_ATTRIBUTES:
            return getattr(self._object, name)
        else:
            raise AttributeError('Attribute "{}" is not allowed to be accessed on related object.'.format(name))

    def __str__(self):
        return self._object.__str__()


class DeletedRelatedObject:
    """
    Placeholder class that substitutes deleted related object and returns:
        * "[DELETED]" as its string representation
        * itself for any attribute accessed
    """
    def __getattr__(self, name):
        return self

    def __str__(self):
        return _('[DELETED]')


class SignalMap:
    """
    Maps signals to arbitrary values.
    """
    def __init__(self):
        self.map = {}

    def add(self, signal, value):
        if signal not in self.map:
            self.map[signal] = [value]
        else:
            self.map[signal].append(value)

    def get(self, signal):
        return self.map.get(signal, [])

    def remove(self, signal):
        if signal in self.map:
            del self.map[signal]


signal_map = SignalMap()


def register(signal, handler_class, allowed_senders=None):
    """
    Starts listening to ``signal`` and registers ``handler_class`` to it.
    """
    signal_map.add(signal, (handler_class, allowed_senders))
    signal.connect(receive, dispatch_uid='pynotify')


def receive(sender, **kwargs):
    """
    Initiates processing of the signal by notification handlers through a receiver.
    """
    if not settings.ENABLED:
        return

    signal = kwargs.pop('signal')
    receiver_class = locate(settings.RECEIVER)

    if not receiver_class:
        raise ImproperlyConfigured('Unable to locate receiver class {}'.format(settings.RECEIVER))

    for handler_class, allowed_senders in signal_map.get(signal):
        if allowed_senders and sender not in allowed_senders:
            continue
        receiver_class(handler_class).receive(kwargs)


def autoload():
    """
    Attempts to load (import) notification handlers from modules defined in ``PYNOTIFY_AUTOLOAD_MODULES``
    """
    modules = settings.AUTOLOAD_MODULES
    if modules:
        for module in modules:
            try:
                import_module(module)
            except ImportError:
                logger.exception('Failed to autoload notification handlers from module {}'.format(module))


def process_task(handler_class, serializer_class, signal_kwargs):
    """
    Deserializes signal kwargs using the given serializer and calls given handler.  This function is intended to be
    called from a Celery task.
    """
    handler_class = locate(handler_class)
    serializer_class = locate(serializer_class)
    signal_kwargs = serializer_class().deserialize(signal_kwargs)

    handler_class().handle(signal_kwargs)


def get_import_path(_class):
    """
    Returns import path for a given class.
    """
    return '{}.{}'.format(_class.__module__, _class.__name__)


def get_from_context(variable, context):
    """
    Tries to find `variable` value in given `context`.

    Args:
        variable: Variable to look for. Template format is supported (e.g. "abc.def.ghi").
        context: Template context.

    Returns:
        Variable value or None if not found.
    """
    value = context

    for chunk in variable.split('.'):
        try:
            value = getattr(value, chunk)
        except AttributeError:
            try:
                value = value.get(chunk)
            except AttributeError:
                return None

        if value is None:
            return None

    return value


def strip_html(value):
    """
    Strips HTML (tags and entities) from string `value`.
    """
    # The space is added to remove BS warnings because value "http://django.pynotify.com"
    # will be considered as URL not as string in BS. The space will be removed with get_text method.
    return BeautifulSoup(' ' + value, 'lxml').get_text()



import json
import re

from chamber.models import SmartModel, SmartModelBase, SmartQuerySet
from django.conf import settings as django_settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models
from django.template import Context, Template
from django.utils.functional import cached_property
from django.utils.translation import ugettext as _
from django.utils.translation import ugettext_lazy as _l

from .config import settings
from .exceptions import MissingContextVariableError
from .helpers import DeletedRelatedObject, SecureRelatedObject, get_from_context, strip_html


class BaseModel(SmartModel):
    """
    Base class for models that outpus its verbose name and PK.
    """
    def __str__(self):
        return '{} #{}'.format(self._meta.verbose_name, self.pk)

    class Meta:
        abstract = True


class BaseTemplate(BaseModel):
    """
    Base abstract model for notification template.

    Attributes:
        title: Title of the notification.
        text: Text of the notification.
        trigger_action: Arbitrary action performed when user triggers (i.e. clicks/taps) the notification.
    """
    title = models.CharField(max_length=200, verbose_name=_l('title'))
    text = models.TextField(null=True, blank=True, verbose_name=_l('text'))
    trigger_action = models.CharField(max_length=2500, null=True, blank=True,
                                      verbose_name=_l('trigger action'))

    class Meta:
        abstract = True


class AdminNotificationTemplate(BaseTemplate):
    """
    Represents a "template of a template". This model is intended to be managed from administration, hence its name. It
    is identified by `slug`, which can be used for notification creation. However, this template is never used to
    directly render a notification, but instead is used to create `NotificationTemplate` with same values.

    Attributes:
        slug: Template slug, with which this template can be referred to.
        is_active: Flag that switches on/off creating notifications from this template.
        send_push: Flag that switches on/off sending push notifications from this template.
            Currently, it has no effect on its own, but you can use it in your custom push notification solution.
    """
    slug = models.SlugField(max_length=200, unique=True, verbose_name=_l('slug'))
    is_active = models.BooleanField(default=True, verbose_name=_l('is active'))
    send_push = models.BooleanField(default=False, verbose_name=_l('send push notification'))

    class Meta:
        verbose_name = _l('admin notification template')
        verbose_name_plural = _l('admin notification templates')

    def __str__(self):
        return '{} ({})'.format(super().__str__(), self.slug)


class NotificationTemplate(BaseTemplate):
    """
    Represents template that is used for rendering notification fields. Each field specified in ``TEMPLATE_FIELDS`` is a
    template string, that can be rendered using the ``render`` method.

    Attributes:
        title: Title of the notification.
        text: Text of the notification.
        trigger_action: Arbitrary action performed when user triggers (i.e. clicks/taps) the notification.
        admin_template: Reference to admin template that was used to create this notification template.
    """
    TEMPLATE_FIELDS = ['title', 'text', 'trigger_action']

    admin_template = models.ForeignKey(
        AdminNotificationTemplate,
        related_name="notification_templates",
        on_delete=models.SET_NULL,
        verbose_name=_l('admin notification template'),
        blank=True,
        null=True,
    )

    class Meta:
        verbose_name = _l('notification template')
        verbose_name_plural = _l('notification templates')

    @property
    def slug(self):
        return self.admin_template.slug if self.admin_template else None

    def render(self, field, context):
        """
        Renders ``field`` using ``context``.
        """
        template_string = getattr(self, field) or ''

        if settings.TEMPLATE_TRANSLATE:
            template_string = _(template_string)

        if settings.TEMPLATE_CHECK:
            vars = re.findall(r'{{ ?([^\|} ]+)[^}]*}}', template_string)
            for var in vars:
                value = get_from_context(var, context)
                if value is None or isinstance(value, DeletedRelatedObject):
                    raise MissingContextVariableError(field, var)

        output = Template('{}{}'.format(settings.TEMPLATE_PREFIX, template_string)).render(Context(context))

        if settings.STRIP_HTML:
            output = strip_html(output)

        return output


class NotificationQuerySet(SmartQuerySet):

    def _create_related_object(self, notification, obj, name=None):
        if not isinstance(obj, models.Model):
            raise TypeError('Related object must be an instance of model.')
        NotificationRelatedObject.objects.create(name=name, notification=notification, content_object=obj)

    def create(self, recipient, template, related_objects=None, extra_data=None, **kwargs):
        notification = super().create(recipient=recipient, template=template, **kwargs)

        if related_objects is not None:
            if isinstance(related_objects, dict):
                for name, obj in related_objects.items():
                    self._create_related_object(notification, obj, name)
            elif isinstance(related_objects, list):
                for obj in related_objects:
                    self._create_related_object(notification, obj)
            else:
                raise TypeError('Related objects must be a list or dictionary in form {"name": object}.')

        if extra_data is not None:
            notification.set_extra_data(extra_data)
            notification.save()

        return notification

    def filter_with_related_object(self, related_object):
        return self.filter(
            related_objects__content_type=ContentType.objects.get_for_model(related_object),
            related_objects__object_id=str(related_object.pk),
        )


class NotificationMeta(SmartModelBase, type):
    """
    Creates property for each template field. The property returns rendered template.
    """
    def __new__(cls, name, bases, attrs):
        new_class = super().__new__(cls, name, bases, attrs)
        for field in NotificationTemplate.TEMPLATE_FIELDS:
            setattr(
                new_class,
                field,
                property(lambda self, field=field: self._render(field))
            )

        return new_class


class Notification(BaseModel, metaclass=NotificationMeta):
    """
    Represents the notification.

    Attributes specified in ``NotificationTemplate.TEMPLATE_FIELDS`` are also available here, as generated properties,
    that are evaluated at runtime and will return rendered field from the associated template. By default, the context
    used for rendering is filled with named related objects and extra data, so they can be referenced in the template by
    their name/key.

    Attributes:
        recipient: Recipient of the notification.
        template: Template used to render generated notification fields.
        is_read: Boolean flag indicating that recipitent has seen the notification.
        is_triggered: Boolean flag indicating that recipient has triggered the notification (e.g. clicked/tapped)
        extra_data: JSON serialized dictionary with extra data.
    """
    recipient = models.ForeignKey(
        django_settings.AUTH_USER_MODEL,
        related_name='notifications',
        on_delete=models.CASCADE,
        verbose_name=_l('recipient'),
    )
    template = models.ForeignKey(
        NotificationTemplate,
        related_name="notifications",
        on_delete=models.PROTECT,
        verbose_name=_l('template')
    )
    is_read = models.BooleanField(default=False, verbose_name=_l('is read'))
    is_triggered = models.BooleanField(default=False, verbose_name=_l('is triggered'))
    extra_data = models.TextField(null=True, blank=True, verbose_name=_l('extra data'))

    objects = NotificationQuerySet.as_manager()

    class Meta:
        verbose_name = _l('notification')
        verbose_name_plural = _l('notifications')
        ordering = ('-created_at',)

    def _render(self, field):
        return self.template.render(field, self.context)

    def _pre_save(self, *args, **kwargs):
        keys = set(self.get_extra_data()) & set(obj.name for obj in self.related_objects.all() if obj.name)
        if keys:
            raise ValueError('Related objects and extra data contain same key(s): {}'.format(', '.join(keys)))

    @cached_property
    def related_objects_dict(self):
        """
        Returns named related objects as a dictionary where key is name of the related object and value is the object
        itself. Related objects without name are skipped.
        """
        output = {}
        for obj in self.related_objects.filter(name__isnull=False):
            output[obj.name] = SecureRelatedObject(obj.content_object) if obj.content_object else DeletedRelatedObject()
        return output

    @property
    def context(self):
        """
        Returns context dictionary used for rendering the template.
        """
        return {**self.related_objects_dict, **self.get_extra_data()}

    def set_extra_data(self, extra_data):
        """
        Setter for ``extra_data`` field.

        Arguments:
            extra_data: Dictionary of JSON serializable values.
        """
        if not isinstance(extra_data, dict):
            raise ValueError('Extra data must be a dictionary.')
        self.extra_data = json.dumps(extra_data, cls=DjangoJSONEncoder)

    def get_extra_data(self):
        """
        Getter for ``extra_data`` field.

        Returns:
            Dictionary with extra data.
        """
        return json.loads(self.extra_data) if self.extra_data is not None else {}


class NotificationRelatedObject(BaseModel):
    """
    Represents object related to a notification. This object can be then referenced in notification template
    fields by its `name` (if not None).

    Attributes:
        name: String identificator of the object (for referencing in templates).
        notification: Related notification.
        content_object: The related object itself.
    """
    name = models.CharField(max_length=200, verbose_name=_l('name'), blank=True, null=True)
    notification = models.ForeignKey(
        Notification,
        on_delete=models.CASCADE,
        related_name='related_objects',
        verbose_name=_l('notification'),
    )
    content_type = models.ForeignKey(ContentType, null=True, on_delete=models.SET_NULL, verbose_name=_l('content type'))
    object_id = models.TextField(db_index=True, verbose_name=_l('object ID'))
    content_object = GenericForeignKey('content_type', 'object_id')



# Generated by Django 2.2.15 on 2020-12-07 08:22

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pynotify', '0005_adminnotificationtemplate_is_active'),
    ]

    operations = [
        migrations.AddField(
            model_name='adminnotificationtemplate',
            name='send_push',
            field=models.BooleanField(default=False, verbose_name='send push notification'),
        ),
    ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'beautifulsoup4 ~=4.8.0',
    'celery >= 4.2.0',
    'django >= 2.2',
    'django-chamber >= 0.5.0',
    'lxml >= 4.6.2',
]

setup(
    author="Ondřej Kulatý",
    author_email='kulaty.o@gmail.com',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="General purpose notification library for Django",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='django-pynotify',
    name='django-pynotify',
    packages=find_packages(include=['pynotify']),
    url='https://github.com/druids/django-pynotify',
    version='0.4.5',
    zip_safe=False,
)
